==> graphgallery-0.7.2.tar.gz: graphgallery-0.7.2/graphgallery/functional/sparse/gdc.py <==
import numpy as np
import scipy.sparse as sp
from sklearn.preprocessing import normalize
from scipy.linalg import expm

from .normalize_adj import normalize_adj
from ..transforms import Transform
from ..decorators import multiple
from ..get_transform import Transformers


@Transformers.register()
class GDC(Transform):
    def __init__(self,
                 alpha: float = 0.3,
                 t: float = None,
                 eps: float = None,
                 k: int = 128,
                 which: str = 'PPR'):
        super().__init__()
        self.alpha = alpha
        self.t = t
        self.eps = eps
        self.k = k
        self.which = which

    def __call__(self, adj_matrix):
        return gdc(adj_matrix,
                   alpha=self.alpha,
                   t=self.t,
                   eps=self.eps,
                   k=self.k,
                   which=self.which)

    def extra_repr(self):
        return f"alpha={self.alpha}, t={self.t}, eps={self.eps}, k={self.k}, which={self.which}"


@multiple()
def gdc(adj_matrix: sp.csr_matrix,
        alpha: float = 0.3,
        t: float = None,
        eps: float = None,
        k: int = 128,
        which: str = 'PPR') -> sp.csr_matrix:

    if not (eps or k):
        raise RuntimeError('Either `eps` or `k` should be specified!')
    if eps and k:
        raise RuntimeError('Only one of `eps` and `k` should be specified!')

    N = adj_matrix.shape[0]

    # Symmetric transition matrix
    T_sym = normalize_adj(adj_matrix)

    if which == 'PPR':
        # PPR-based diffusion
        assert alpha, '`alpha` should be specified for PPR-based diffusion.'
        S = alpha * sp.linalg.inv((sp.eye(N) - (1 - alpha) * T_sym).tocsc())
    elif which == 'Heat':
        assert t, '`t` should be specified for Heat-based diffusion.'
        S = -t * (sp.eye(N) - T_sym)
        S = expm(S.toarray())
        S = sp.csr_matrix(S)
    else:
        raise ValueError(f'Invalid argument of `{which}`.')

    if eps:
        # Sparsify using threshold epsilon
        S = clip_matrix(S, threshold=eps)
    else:
        # Row-wise select top-k values
        S = top_k_matrix(S, k=k)

    # Column-normalized transition matrix on graph S_tilde
    T_S = normalize(S, norm='l1', axis=0)

    return T_S.tocsr(copy=False)


def clip_matrix(matrix, threshold: float) -> sp.csr_matrix:
    '''Sparsify using threshold epsilon'''
    assert sp.isspmatrix(
        matrix
    ), 'Input matrix should be sparse matrix with format scipy.sparse.*_matrix.'
    matrix = matrix.tocsr()
    thres = np.vectorize(lambda x: x if x >= threshold else 0.)
    matrix.data = thres(matrix.data)
    matrix.eliminate_zeros()
    return matrix


def top_k_matrix(matrix, k: int) -> sp.csr_matrix:
    '''Row-wise select top-k values'''
    assert sp.isspmatrix(
        matrix
    ), 'Input matrix should be sparse matrix with format scipy.sparse.*_matrix.'
    matrix = matrix.tolil()
    data = matrix.data
    for row in range(matrix.shape[0]):
        t = np.asarray(data[row])
        t[np.argsort(-t)[k:]] = 0.
        data[row] = t.tolist()
    matrix = matrix.tocsr()
    matrix.eliminate_zeros()
    return matrix

==> graphgallery-1.0.0.tar.gz: graphgallery-1.0.0/graphgallery/functional/sparse/gdc.py <==
import scipy.sparse as sp
from sklearn.preprocessing import normalize
from scipy.linalg import expm

from .normalize_adj import normalize_adj
from ..base_transforms import SparseTransform
from ..decorators import multiple
from ..transform import Transform
from .topk import sparse_topk
from .clip import sparse_clip


@Transform.register()
class GDC(SparseTransform):
    def __init__(self,
                 alpha: float = 0.3,
                 t: float = None,
                 eps: float = None,
                 K: int = 128,
                 which: str = 'PPR'):
        super().__init__()
        self.collect(locals())

    def __call__(self, adj_matrix):
        return gdc(adj_matrix,
                   alpha=self.alpha,
                   t=self.t,
                   eps=self.eps,
                   K=self.K,
                   which=self.which)


@multiple()
def gdc(adj_matrix: sp.csr_matrix,
        alpha: float = 0.3,
        t: float = None,
        eps: float = None,
        K: int = 128,
        which: str = 'PPR') -> sp.csr_matrix:

    if not (eps or K):
        raise RuntimeError('Either `eps` or `K` should be specified!')
    if eps and K:
        raise RuntimeError('Only one of `eps` and `K` should be specified!')

    N = adj_matrix.shape[0]

    # Symmetric transition matrix
    T_sym = normalize_adj(adj_matrix)

    if which == 'PPR':
        # PPR-based diffusion
        assert alpha, '`alpha` should be specified for PPR-based diffusion.'
        S = alpha * sp.linalg.inv((sp.eye(N, format='csr') - (1 - alpha) * T_sym).tocsc())
    elif which == 'Heat':
        assert t, '`t` should be specified for Heat-based diffusion.'
        S = -t * (sp.eye(N, format='csr') - T_sym)
        S = expm(S.toarray())
        S = sp.csr_matrix(S)
    else:
        raise ValueError(f'Invalid argument of `{which}`.')

    if eps:
        # Sparsify using threshold epsilon
        S = sparse_clip(S, threshold=eps)
    else:
        # Row-wise select top-K values
        S = sparse_topk(S, K=K)

    # Column-normalized transition matrix on graph S_tilde
    T_S = normalize(S, norm='l1', axis=0)

    return T_S.tocsr(copy=False)
